@@ -1,6 +1,6 @@
 ;;; lui.el --- Linewise User Interface -*- lexical-binding: t -*-
 
-;; Copyright (C) 2005 - 2015  Jorgen Schaefer
+;; Copyright (C) 2005 - 2016  Jorgen Schaefer
 
 ;; Author: Jorgen Schaefer <forcer@forcix.cx>
 ;; URL: https://github.com/jorgenschaefer/circe/wiki/Lui
@@ -37,6 +37,8 @@
 ;; lui-insert
 ;; lui-input-function
 ;; lui-completion-function
+;; lui-time-stamp-time
+;; lui-time-stamp-zone
 ;; and the 'lui-fool and 'lui-do-not-track text properties
 
 ;;; Code:
@@ -131,6 +133,7 @@ a valid list of dictionaries."
 
 (defcustom lui-highlight-keywords nil
   "A list of keywords to highlight.
+
 This specifies a list of keywords that Lui should highlight. Each
 entry is of one of the following forms (similar to
 `font-lock-keywords'):
@@ -146,7 +149,10 @@ entry is of one of the following forms (similar to
     Highlight the SUBMATCH in REGEXP in FACE
 
 In all of these cases, the FACE can also be a property list which
-is then associated with the match."
+is then associated with the match.
+
+All matches are run, which means later matches can override
+changes by earlier ones."
   :type '(repeat (choice
                   (string :tag "Regular Expression")
                   (list :tag "Submatch"
@@ -161,6 +167,27 @@ is then associated with the match."
                         (face :tag "Face"))))
   :group 'lui)
 
+(defface lui-strong-face
+  '((t (:inherit bold)))
+  "Face used for strong markup."
+  :group 'lui-irc-colors)
+
+(defface lui-emphasis-face
+  '((t (:inherit italic)))
+  "Face for emphasis markup."
+  :group 'lui-irc-colors)
+
+(defcustom lui-formatting-list nil
+  "List of enabled formatting types.
+Each list item is a list consisting of a regular expression
+matching the highlighted text, an integer for the submatch and a
+face for highlighting the match."
+  :type '(set (const :tag "*Strong* text"
+                     ("\\*\\([^*]*\\)\\*" 1 lui-strong-face))
+              (const :tag "_Emphasized_ text"
+                     ("_\\([^_]*\\)_" 1 lui-emphasis-face)))
+  :group 'lui)
+
 (defcustom lui-buttons-list
   `(("`\\([A-Za-z0-9+=*/-]+\\)'" 1
      lui-button-elisp-symbol 1)
@@ -478,17 +505,21 @@ This is called from `post-command-hook'.
 
 See `lui-scroll-behavior' for how to customize this."
   (condition-case err
-      (when (and lui-input-marker
-                 (memq lui-scroll-behavior '(t post-command)))
-        ;; Code from ERC's erc-goodies.el. I think this was originally
-        ;; mine anyhow, not sure though.
-        (save-restriction
-          (when (>= (point) lui-input-marker)
-            (save-excursion
-              (goto-char (point-max))
-              (recenter -1)))))
+      (dolist (w (window-list))
+        (with-current-buffer (window-buffer w)
+          (when (and lui-input-marker
+                     (memq lui-scroll-behavior '(t post-command)))
+            ;; Code from ERC's erc-goodies.el. I think this was originally
+            ;; mine anyhow, not sure though.
+            (save-restriction
+              (widen)
+              (when (>= (point) lui-input-marker)
+                (save-excursion
+                  (goto-char (point-max))
+                  (with-selected-window w
+                    (recenter -1))))))))
     (error
-     (message "Error in l-s-p-c: %S" err)
+     (message "Error in lui-scroll-post-command: %S" err)
      )))
 
 (defun lui-scroll-post-output ()
@@ -524,15 +555,15 @@ If point is not in the input area, insert a newline."
       (run-hooks 'lui-pre-input-hook))
     (let ((input (buffer-substring lui-input-marker (point-max))))
       (delete-region lui-input-marker (point-max))
-      (ring-insert lui-input-ring input)
-      (setq lui-input-ring-index nil)
+      (lui-add-input input)
       (if lui-input-function
           (funcall lui-input-function input)
-        (error "No input function specified")))
-    ;; Clean the undo list. The input is gone, nothing to be undone
-    ;; here.
-    ;     (setq buffer-undo-list nil)
-    ))
+        (error "No input function specified")))))
+
+(defun lui-add-input (input)
+  "Add INPUT as if entered by the user."
+  (ring-insert lui-input-ring input)
+  (setq lui-input-ring-index nil))
 
 
 ;;;;;;;;;;;;;;;
@@ -891,6 +922,7 @@ of the buffer."
                               (point-max)
                               `(lui-raw-text ,str))
          (run-hooks 'lui-pre-output-hook)
+         (lui-apply-formatting)
          (lui-highlight-keywords)
          (lui-buttonize)
          (lui-fill)
@@ -1075,6 +1107,18 @@ This is called automatically when new text is inserted."
                                  (append (funcall properties entry)
                                          '(lui-highlight-fontified-p t)))))))))
 
+(defun lui-apply-formatting ()
+  "Highlight the entries in `lui-formatting-list'."
+  (dolist (entry lui-formatting-list)
+    (goto-char (point-min))
+    (let ((re (car entry))
+          (subgroup (cadr entry))
+          (face (nth 2 entry)))
+      (while (re-search-forward re nil t)
+        (when face
+          (add-face-text-property (match-beginning subgroup) (match-end subgroup)
+                                  face nil (current-buffer)))))))
+
 
 ;;;;;;;;;;;;;;;
 ;;; Filling ;;;
@@ -1135,9 +1179,28 @@ function."
   "The last time stamp.")
 (make-variable-buffer-local 'lui-time-stamp-last)
 
+(defvar lui-time-stamp-time nil
+  "A custom time to use as the time stamp for `lui-insert'.
+
+This variable should be let-bound when you wish to provide a
+custom time to be printed by `lui-time-stamp'.  If this variable
+is nil the current time is used.  See the TIME argument to
+`format-time-string' for more information.")
+
+(defvar lui-time-stamp-zone nil
+  "A custom timezone to use for the time stamp for `lui-insert'.
+
+This variable should be let-bound when you wish to provide a
+custom time zone when printing the time stamp with
+`lui-time-stamp'.  If this variable is nil local time is used.
+See the ZONE argument to `format-time-string' for more
+information.")
+
 (defun lui-time-stamp ()
   "Add a time stamp to the current buffer."
-  (let ((ts (format-time-string lui-time-stamp-format)))
+  (let ((ts (format-time-string lui-time-stamp-format
+                                lui-time-stamp-time
+                                lui-time-stamp-zone)))
     (cond
      ;; Time stamps right
      ((or (numberp lui-time-stamp-position)
